from django.shortcuts import render, redirect
from django.http import JsonResponse
from django.conf import settings
import time
import os
import rust_crud_api  # This is the module we built

db_url = settings.DATABASE_URL

def init_db_view(request):
    db_url = os.environ.get("DATABASE_URL")
    try:
        rust_crud_api.init_db(db_url)
        return JsonResponse({"message": "Database initialized"})
    except Exception as e:
        return JsonResponse({"error": str(e)}, status=500)

def register_view(request):
    """
    Handles user registration. Expects POST with 'name', 'email', and 'password'.
    Calls the Rust extension to create the user and then redirects to the login page.
    """
    db_url = settings.DATABASE_URL  # Defined in settings.py
    context = {}
    
    if request.method == 'POST':
        name = request.POST.get('name', '').strip()
        email = request.POST.get('email', '').strip()
        password = request.POST.get('password', '').strip()
        
        if not name or not email or not password:
            context['error'] = "Name, email, and password are required."
            return render(request, 'myapp/register.html', context)
        
        try:
            # Create the user with the hashed password via the Rust extension.
            rust_crud_api.create_user(db_url, name, email, password)
            # Redirect to the login page after successful registration.
            return redirect('login')
        except Exception as e:
            context['error'] = f"An error occurred: {str(e)}"
            return render(request, 'myapp/register.html', context)
    
    return render(request, 'myapp/register.html', context)


def login_view(request):
    """
    Handles user login by verifying credentials using the Rust extension.
    If valid, stores user details in the session and redirects to the account page.
    """
    db_url = settings.DATABASE_URL
    context = {}
    
    if request.method == 'POST':
        email = request.POST.get('email', '').strip()
        password = request.POST.get('password', '').strip()
        
        if not email or not password:
            context['error'] = "Email and password are required."
            return render(request, 'myapp/login.html', context)
        
        try:
            # Verify the user's credentials using the Rust extension.
            is_valid = rust_crud_api.verify_user(db_url, email, password)
            if not is_valid:
                context['error'] = "Invalid email or password."
                return render(request, 'myapp/login.html', context)
            
            # Retrieve the user record to store additional info in session.
            # (For simplicity, we retrieve all users and select the matching one.)
            users = rust_crud_api.get_all_users(db_url)
            user = next((u for u in users if u.email.lower() == email.lower()), None)
            if not user:
                context['error'] = "User not found."
                return render(request, 'myapp/login.html', context)
            
            # Store user information in the session.
            request.session['user_id'] = user.id
            request.session['user_name'] = user.name
            request.session['user_email'] = user.email
            
            return redirect('account')
        except Exception as e:
            context['error'] = f"An error occurred: {str(e)}"
            return render(request, 'myapp/login.html', context)
    
    return render(request, 'myapp/login.html', context)

def account_view(request):
    """
    Display the account page for the logged-in user.
    """
    user_id = request.session.get('user_id')
    if not user_id:
        return redirect('login')
    db_url = settings.DATABASE_URL
    try:
        user = rust_crud_api.get_user(db_url, int(user_id))
        if user is None:
            return JsonResponse({'error': 'User not found.'}, status=404)
        return render(request, 'myapp/account.html', {'user': user})
    except Exception as e:
        return JsonResponse({'error': str(e)}, status=500)


def delete_account_view(request):
    """
    Delete the account of the logged-in user.
    """
    user_id = request.session.get('user_id')
    if not user_id:
        return redirect('login')
    db_url = settings.DATABASE_URL
    if request.method == 'POST':
        try:
            success = rust_crud_api.delete_user(db_url, int(user_id))
            if success:
                request.session.flush()  # Clear the session after deletion.
                return redirect('register')
            else:
                return JsonResponse({'error': 'Account deletion failed.'}, status=500)
        except Exception as e:
            return JsonResponse({'error': str(e)}, status=500)
    return render(request, 'myapp/delete_account.html')

def logout_view(request):
    # Clear all session data to log out the user.
    request.session.flush()
    # Optionally, render a confirmation page or redirect to login.
    return render(request, 'myapp/logout.html')

