/// lib.rs

use pyo3::prelude::*;
use pyo3::exceptions::PyRuntimeError;
use postgres::{Client, NoTls};
use pyo3::types::PyDict;
use serde::{Serialize, Deserialize};  /// Import derive macros from serde_derive
use jsonwebtoken::{encode, decode, Header, Validation, EncodingKey, DecodingKey};


/// Define our User model, which will be exposed to Python.
#[pyclass]
#[derive(Serialize, Deserialize, Debug)]
struct User {
    #[pyo3(get, set)]
    id: Option<i32>,
    #[pyo3(get, set)]
    name: String,
    #[pyo3(get, set)]
    email: String,
}
/// Group Model
#[pyclass]
#[derive(Serialize, Deserialize, Debug)]
struct Group {
    #[pyo3(get, set)]
    id: Option<i32>,
    #[pyo3(get, set)]
    name: String,
}

//JWT Claims for Token Payload

#[derive(Debug,Serialize,Deserialize)]
struct Claims {
    sub: String, // Subject 
    exp: usize, // Expiration time
}

/// Generate a JWT token for a given user.
/// Returns the token as a String.
#[pyfunction]
fn generate_jwt(user_email: String, secret: String, exp: usize) -> PyResult<String> {
    let claims = Claims { sub: user_email, exp };
    encode(&Header::default(), &claims, &EncodingKey::from_secret(secret.as_ref()))
        .map_err(|e| PyRuntimeError::new_err(e.to_string()))
}

/// Verify a JWT token and return its claims as a Python dictionary.
#[pyfunction]
fn verify_jwt(token: String, secret: String) -> PyResult<PyObject> {
    let gil = Python::acquire_gil();
    let py = gil.python();
    
    let token_data = decode::<Claims>(
        &token,
        &DecodingKey::from_secret(secret.as_ref()),
        &Validation::default()
    ).map_err(|e| PyRuntimeError::new_err(e.to_string()))?;
    
    // Extract the claims.
    let claims = token_data.claims;
    // Create a Python dictionary and insert the claim fields.
    let dict = PyDict::new(py);
    dict.set_item("sub", claims.sub)?;
    dict.set_item("exp", claims.exp)?;
    Ok(dict.to_object(py))
}

// Pasword Hashing with Argon2

#[pyfunction]
fn hash_password(password: &str) -> PyResult<String> {
    use argon2::{Argon2, PasswordHasher};
    use argon2::password_hash::{SaltString, rand_core::OsRng};
    
    // Generate a random salt using OsRng.
    let salt = SaltString::generate(&mut OsRng);
    // Create a default Argon2 instance.
    let argon2 = Argon2::default();
    // Hash the password using the salt.
    let password_hash = argon2.hash_password(password.as_bytes(), &salt)
        .map_err(|e| PyRuntimeError::new_err(e.to_string()))?
        .to_string();
    Ok(password_hash)
}

#[pyfunction]
fn verify_password(hash: &str, password: &str) -> PyResult<bool> {
    use argon2::{Argon2, PasswordVerifier};
    use argon2::password_hash::PasswordHash;
    
    let argon2 = Argon2::default();
    let parsed_hash = PasswordHash::new(hash)
        .map_err(|e| PyRuntimeError::new_err(e.to_string()))?;
    argon2.verify_password(password.as_bytes(), &parsed_hash)
        .map_err(|e| PyRuntimeError::new_err(e.to_string()))
        .map(|_| true)
}

/// A helper function to convert postgres::Error into a Python RuntimeError.
fn pg_err(e: postgres::Error) -> PyErr {
    PyRuntimeError::new_err(e.to_string())
}

/// Initialize the database by creating the users table if it does not exist.
#[pyfunction]
fn init_db(db_url: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.batch_execute(
        "
        CREATE TABLE IF NOT EXISTS users (
            id SERIAL PRIMARY KEY,
            name VARCHAR NOT NULL,
            email VARCHAR NOT NULL UNIQUE,
            password_hash VARCHAR NOT NULL
        );
        CREATE TABLE IF NOT EXISTS groups (
            id SERIAL PRIMARY KEY,
            name VARCHAR NOT NULL
        );
        CREATE TABLE IF NOT EXISTS group_members (
            group_id INTEGER REFERENCES groups(id),
            user_id INTEGER REFERENCES users(id),
            PRIMARY KEY (group_id, user_id)
        );
        "
    ).map_err(pg_err)?;
    Ok(())
}

/// Create a new user by inserting into the database.
#[pyfunction]
fn create_user(db_url: &str, name: &str, email: &str, password: &str) -> PyResult<()> {
    let password_hash = hash_password(password)?;
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO users (name, email, password_hash) VALUES ($1, $2, $3)",
        &[&name, &email, &password_hash]
    ).map_err(pg_err)?;
    Ok(())
}

/// Retrieve a user by ID. Returns None if the user is not found.
#[pyfunction]
fn get_user(db_url: &str, user_id: i32) -> PyResult<Option<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT id, name, email FROM users WHERE id = $1",
        &[&user_id]
    ).map_err(pg_err)?;
    
    if let Some(row) = row_opt {
        let user = User {
            id: row.get(0),
            name: row.get(1),
            email: row.get(2),
        };
        Ok(Some(user))
    } else {
        Ok(None)
    }
}

/// Retrieve all users from the database.
#[pyfunction]
fn get_all_users(db_url: &str) -> PyResult<Vec<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query("SELECT id, name, email FROM users", &[])
        .map_err(pg_err)?;
    
    let mut users = Vec::new();
    for row in rows {
        users.push(User {
            id: row.get(0),
            name: row.get(1),
            email: row.get(2),
        });
    }
    Ok(users)
}

/// Update an existing user with a new name and email.
#[pyfunction]
fn update_user(db_url: &str, user_id: i32, new_name: &str, new_email: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let updated = client.execute(
        "UPDATE users SET name = $1, email = $2 WHERE id = $3",
        &[&new_name, &new_email, &user_id]
    ).map_err(pg_err)?;
    if updated == 0 {
        Err(PyRuntimeError::new_err("User not found"))
    } else {
        Ok(())
    }
}

/// Delete a user by ID. Returns true if a user was deleted.
#[pyfunction]
fn delete_user(db_url: &str, user_id: i32) -> PyResult<bool> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let deleted = client.execute(
        "DELETE FROM users WHERE id = $1",
        &[&user_id]
    ).map_err(pg_err)?;
    Ok(deleted > 0)
}

///Verify credentials
///Retrieve a stored password hash for the email and compare.
#[pyfunction]
fn verify_user(db_url: &str, email: &str, password: & str) -> PyResult<bool> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT password_hash FROM users WHERE email = $1",
        &[&email]
        ).map_err(pg_err)?;
        if let Some(row) = row_opt {
            let stored_hash: String = row.get(0);
            verify_password(&stored_hash, password)
            } else {
                Ok(false)
        }
}


/// Create a group
#[pyfunction]
fn create_group(db_url: &str, name: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO groups (name) VALUES ($1)",
        &[&name]
    ).map_err(pg_err)?;
    Ok(())
}

/// Retrieve a group by ID.
#[pyfunction]
fn get_group(db_url: &str, group_id: i32) -> PyResult<Option<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT id, name FROM groups WHERE id = $1",
        &[&group_id]
    ).map_err(pg_err)?;
    
    if let Some(row) = row_opt {
        let group = Group {
            id: row.get(0),
            name: row.get(1),
        };
        Ok(Some(group))
    } else {
        Ok(None)
    }
}

/// Retrieve all groups.
#[pyfunction]
fn get_all_groups(db_url: &str) -> PyResult<Vec<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query("SELECT id, name FROM groups", &[])
        .map_err(pg_err)?;
    
    let groups = rows.into_iter().map(|row| Group {
        id: row.get(0),
        name: row.get(1),
    }).collect();
    
    Ok(groups)
}

/// Add a user to a group.
#[pyfunction]
fn add_user_to_group(db_url: &str, group_id: i32, user_id: i32) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO group_members (group_id, user_id) VALUES ($1, $2)",
        &[&group_id, &user_id]
    ).map_err(pg_err)?;
    Ok(())
}

/// Get all members of a group.
#[pyfunction]
fn get_group_members(db_url: &str, group_id: i32) -> PyResult<Vec<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query(
        "SELECT u.id, u.name, u.email
         FROM users u
         JOIN group_members gm ON u.id = gm.user_id
         WHERE gm.group_id = $1",
         &[&group_id]
    ).map_err(pg_err)?;
    
    let users = rows.into_iter().map(|row| User {
        id: row.get(0),
        name: row.get(1),
        email: row.get(2),
    }).collect();
    
    Ok(users)
}


/// This is the Python module initializer. The name "rust_crud_api" here should match the name used in your Cargo.toml.
#[pymodule]
fn rust_crud_api(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_class::<User>()?;
    m.add_class::<Group>()?;
    m.add_function(wrap_pyfunction!(init_db, m)?)?;
    m.add_function(wrap_pyfunction!(create_user, m)?)?;
    m.add_function(wrap_pyfunction!(get_user, m)?)?;
    m.add_function(wrap_pyfunction!(get_all_users, m)?)?;
    m.add_function(wrap_pyfunction!(update_user, m)?)?;
    m.add_function(wrap_pyfunction!(delete_user, m)?)?;
    m.add_function(wrap_pyfunction!(create_group, m)?)?;
    m.add_function(wrap_pyfunction!(get_group, m)?)?;
    m.add_function(wrap_pyfunction!(get_all_groups, m)?)?;
    m.add_function(wrap_pyfunction!(add_user_to_group, m)?)?;
    m.add_function(wrap_pyfunction!(get_group_members, m)?)?;
    m.add_function(wrap_pyfunction!(generate_jwt, m)?)?;
    m.add_function(wrap_pyfunction!(verify_jwt, m)?)?;
    m.add_function(wrap_pyfunction!(hash_password, m)?)?;
    m.add_function(wrap_pyfunction!(verify_password, m)?)?;
    m.add_function(wrap_pyfunction!(verify_user, m)?)?;
    Ok(())
}

